/*
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

extern "C" {

#include "libavutil/opt.h"
#include "libavutil/imgutils.h"
#include "avfilter.h"
#include "formats.h"
#include "internal.h"
#include "video.h"

}

extern "C" {

typedef struct ColorizeLSDContext {
    const AVClass *classs;

    // user input
    float speed;
    float hue;
    float saturation;
    float lightness;
    float mix;
    float mixcolor;

    int depth; // bit depth of planes
    int c[3]; // yuv of current overlay color (recalculated every frame)

    // size of input/output planes
    int planewidth[4];
    int planeheight[4];

    // ptr to multithreaded functions
    // one function for the y plane one for uv
    int (*do_plane_slice[2])(AVFilterContext *s, void *arg,
                             int jobnr, int nb_jobs, int plane);
} ColorizeLSDContext;

} // extern c

static inline float lerpf(float v0, float v1, float f)
{
    return v0 + (v1 - v0) * f;
}


// luminance slices
template<typename T, bool color>
static int colorize_slice(AVFilterContext *ctx, void *arg, int jobnr, int nb_jobs, int plane)
{
    ColorizeLSDContext *s = (ColorizeLSDContext *) ctx->priv;
    AVFrame *frame = (AVFrame *) arg;
    const int width = s->planewidth[color]; // plane 1 & 2 are the same size
    const int height = s->planeheight[color];
    const int slice_start = (height * jobnr) / nb_jobs;
    const int slice_end = (height * (jobnr + 1)) / nb_jobs;

    const float mix = color ? s->mixcolor : s->mix;

    const int linesize = frame->linesize[plane] / sizeof(T);

    T *ptr = (T*) frame->data[plane] + slice_start * linesize;

    const int c = s->c[plane]; // planes color component of target color

    for (int y = slice_start; y < slice_end; y++) {
        for (int x = 0; x < width; x++) {
            ptr[x] = lerpf(c, ptr[x], mix);
        }

        ptr += linesize;
    }

    return 0;
}

extern "C" {

static int do_slice(AVFilterContext *ctx, void *arg, int jobnr, int nb_jobs)
{
    ColorizeLSDContext *s = (ColorizeLSDContext *) ctx->priv;

    s->do_plane_slice[0](ctx, arg, jobnr, nb_jobs, 0); // y plane
    s->do_plane_slice[1](ctx, arg, jobnr, nb_jobs, 1); // u plane
    s->do_plane_slice[1](ctx, arg, jobnr, nb_jobs, 2); // v plane

    return 0;
}

static float hue2rgb(float p, float q, float t)
{
    if (t < 0.f) t += 1.f;
    if (t > 1.f) t -= 1.f;
    if (t < 1.f/6.f) return p + (q - p) * 6.f * t;
    if (t < 1.f/2.f) return q;
    if (t < 2.f/3.f) return p + (q - p) * (2.f/3.f - t) * 6.f;

    return p;
}

static void hsl2rgb(float h, float s, float l, float *r, float *g, float *b)
{
    h /= 360.f;

    if (s == 0.f) {
        *r = *g = *b = l;
    } else {
        const float q = l < 0.5f ? l * (1.f + s) : l + s - l * s;
        const float p = 2.f * l - q;

        *r = hue2rgb(p, q, h + 1.f / 3.f);
        *g = hue2rgb(p, q, h);
        *b = hue2rgb(p, q, h - 1.f / 3.f);
    }
}

static void rgb2yuv(float r, float g, float b, int *y, int *u, int *v, int depth)
{
    *y = ((0.21260*219.0/255.0) * r + (0.71520*219.0/255.0) * g +
         (0.07220*219.0/255.0) * b) * ((1 << depth) - 1);
    *u = (-(0.11457*224.0/255.0) * r - (0.38543*224.0/255.0) * g +
         (0.50000*224.0/255.0) * b + 0.5) * ((1 << depth) - 1);
    *v = ((0.50000*224.0/255.0) * r - (0.45415*224.0/255.0) * g -
         (0.04585*224.0/255.0) * b + 0.5) * ((1 << depth) - 1);
}

static int filter_frame(AVFilterLink *inlink, AVFrame *frame)
{
    double timebase = av_q2d(inlink->time_base);
    AVFilterContext *ctx = (AVFilterContext *) inlink->dst;
    ColorizeLSDContext *s = (ColorizeLSDContext *) ctx->priv;
    float c[3]; // temp rgb

    double time = timebase * frame->pts;
    uint32_t pos = (uint32_t) ((time * s->speed * 360) + s->hue) % 360;

    hsl2rgb(pos, s->saturation, s->lightness, &c[0], &c[1], &c[2]);
    rgb2yuv(c[0], c[1], c[2], &s->c[0], &s->c[1], &s->c[2], s->depth);

    ff_filter_execute(ctx, do_slice, frame, NULL,
                      FFMIN(s->planeheight[1], ff_filter_get_nb_threads(ctx)));

    return ff_filter_frame(ctx->outputs[0], frame);
}

static const enum AVPixelFormat pixel_fmts[] = {
    AV_PIX_FMT_YUV410P, AV_PIX_FMT_YUV411P,
    AV_PIX_FMT_YUV420P, AV_PIX_FMT_YUV422P,
    AV_PIX_FMT_YUV440P, AV_PIX_FMT_YUV444P,
    AV_PIX_FMT_YUVJ420P, AV_PIX_FMT_YUVJ422P,
    AV_PIX_FMT_YUVJ440P, AV_PIX_FMT_YUVJ444P,
    AV_PIX_FMT_YUVJ411P,
    AV_PIX_FMT_YUV420P9, AV_PIX_FMT_YUV422P9, AV_PIX_FMT_YUV444P9,
    AV_PIX_FMT_YUV420P10, AV_PIX_FMT_YUV422P10, AV_PIX_FMT_YUV444P10,
    AV_PIX_FMT_YUV440P10,
    AV_PIX_FMT_YUV444P12, AV_PIX_FMT_YUV422P12, AV_PIX_FMT_YUV420P12,
    AV_PIX_FMT_YUV440P12,
    AV_PIX_FMT_YUV444P14, AV_PIX_FMT_YUV422P14, AV_PIX_FMT_YUV420P14,
    AV_PIX_FMT_YUV420P16, AV_PIX_FMT_YUV422P16, AV_PIX_FMT_YUV444P16,
    AV_PIX_FMT_YUVA420P,  AV_PIX_FMT_YUVA422P,   AV_PIX_FMT_YUVA444P,
    AV_PIX_FMT_YUVA444P9, AV_PIX_FMT_YUVA444P10, AV_PIX_FMT_YUVA444P12, AV_PIX_FMT_YUVA444P16,
    AV_PIX_FMT_YUVA422P9, AV_PIX_FMT_YUVA422P10, AV_PIX_FMT_YUVA422P12, AV_PIX_FMT_YUVA422P16,
    AV_PIX_FMT_YUVA420P9, AV_PIX_FMT_YUVA420P10, AV_PIX_FMT_YUVA420P16,
    AV_PIX_FMT_NONE
};

static av_cold int config_input(AVFilterLink *inlink)
{
    AVFilterContext *ctx = (AVFilterContext *) inlink->dst;
    ColorizeLSDContext *s = (ColorizeLSDContext *) ctx->priv;
    const AVPixFmtDescriptor *desc = av_pix_fmt_desc_get((AVPixelFormat) inlink->format);
    int depth;

    s->depth = depth = desc->comp[0].depth;

    s->planewidth[1] = s->planewidth[2] = AV_CEIL_RSHIFT(inlink->w, desc->log2_chroma_w);
    s->planewidth[0] = s->planewidth[3] = inlink->w;
    s->planeheight[1] = s->planeheight[2] = AV_CEIL_RSHIFT(inlink->h, desc->log2_chroma_h);
    s->planeheight[0] = s->planeheight[3] = inlink->h;

    s->do_plane_slice[0] = depth <= 8 ? colorize_slice<uint8_t, false> : colorize_slice<uint16_t, false>;
    s->do_plane_slice[1] = depth <= 8 ? colorize_slice<uint8_t, true> : colorize_slice<uint16_t, true>;

    return 0;
}

static const AVFilterPad colorizelsd_inputs[] = {
    {
        .name           = "default",
        .type           = AVMEDIA_TYPE_VIDEO,
        .flags          = AVFILTERPAD_FLAG_NEEDS_WRITABLE,
        .filter_frame   = filter_frame,
        .config_props   = config_input,
    },
};

static const AVFilterPad colorizelsd_outputs[] = {
    {
        .name = "default",
        .type = AVMEDIA_TYPE_VIDEO,
    },
};

#define OFFSET(x) offsetof(ColorizeLSDContext, x)
#define VF AV_OPT_FLAG_FILTERING_PARAM|AV_OPT_FLAG_VIDEO_PARAM|AV_OPT_FLAG_RUNTIME_PARAM

static const AVOption colorizelsd_options[] = {
    { "speed",      "set the rotation speed in r/s",   OFFSET(speed), AV_OPT_TYPE_FLOAT, {.dbl=1}, 0.0000001, 1000, VF},
    { "hue",        "set the start hue",               OFFSET(hue),        AV_OPT_TYPE_FLOAT, {.dbl=0},  0, 360, VF },
    { "saturation", "set the start saturation",        OFFSET(saturation), AV_OPT_TYPE_FLOAT, {.dbl=0.5},0,   1, VF },
    { "lightness",  "set the start lightness",         OFFSET(lightness),  AV_OPT_TYPE_FLOAT, {.dbl=0.5},0,   1, VF },
    { "mix",        "set the mix of source lightness", OFFSET(mix),        AV_OPT_TYPE_FLOAT, {.dbl=1},  0,   1, VF },
    { "mixcolor",   "set the mix of source color",     OFFSET(mixcolor),   AV_OPT_TYPE_FLOAT, {.dbl=0.5}, 0,   1, VF },
    { NULL }
};

AVFILTER_DEFINE_CLASS(colorizelsd);

extern "C" const AVFilter ff_vf_colorizelsd = {
    .name          = "colorizelsd",
    .description   = NULL_IF_CONFIG_SMALL("Overlay a color change on the video stream."),
    // FILTER_INPUTS(colorizelsd_inputs),
    // FILTER_OUTPUTS(colorizelsd_outputs),
    .inputs = colorizelsd_inputs,
    .outputs = colorizelsd_outputs,
    .priv_class = &colorizelsd_class,
    .flags         = AVFILTER_FLAG_SUPPORT_TIMELINE_GENERIC | AVFILTER_FLAG_SLICE_THREADS,
    .nb_inputs = 1,
    .nb_outputs = 1,
    .formats_state = FF_FILTER_FORMATS_PIXFMT_LIST,

    // preinit
    // init
    // init_dict
    // uninit

    .formats = { .pixels_list = pixel_fmts },
    .priv_size     = sizeof(ColorizeLSDContext),

    // FILTER_PIXFMTS_ARRAY(pixel_fmts),
    .process_command = NULL, // ff_filter_process_command

    // activate
};
} // extern c